import unittest
from functools import wraps
import numpy as np

import ams

try:
    from ams.shared import igraph
    getattr(igraph, '__version__')
    HAVE_IGRAPH = True
except (ImportError, AttributeError):
    HAVE_IGRAPH = False


def require_igraph(f):
    """
    Decorator for functions that require igraph.
    """

    @wraps(f)
    def wrapper(*args, **kwds):
        try:
            getattr(igraph, '__version__')
        except AttributeError:
            raise ModuleNotFoundError("igraph needs to be manually installed.")

        return f(*args, **kwds)

    return wrapper


class TestRoutineMethods(unittest.TestCase):
    """
    Test methods of `Routine`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("ieee39/ieee39_uced.xlsx"),
                           default_config=True,
                           no_output=True,
                           )

    def test_data_check(self):
        """
        Test `Routine._data_check()` method.
        """

        self.assertTrue(self.ss.DCOPF._data_check())
        self.assertFalse(self.ss.RTEDES._data_check())

    def test_get_off_constrs(self):
        """
        Test `Routine._get_off_constrs()` method.
        """

        self.assertIsInstance(self.ss.DCOPF._get_off_constrs(), list)

    def test_routine_set(self):
        """
        Test `Routine.set()` method.
        """

        self.ss.DCOPF.set('c2', 'GCost_1', 'v', 10)
        np.testing.assert_equal(self.ss.GCost.get('c2', 'GCost_1', 'v'), 10)

    def test_routine_get(self):
        """
        Test `Routine.get()` method.
        """

        # get an rparam value
        np.testing.assert_equal(self.ss.DCOPF.get('ug', 'PV_30'), 1)

        # get an unpacked var value
        self.ss.DCOPF.run(solver='ECOS')
        self.assertEqual(self.ss.DCOPF.exit_code, 0, "Exit code is not 0.")
        np.testing.assert_equal(self.ss.DCOPF.get('pg', 'PV_30', 'v'),
                                self.ss.StaticGen.get('p', 'PV_30', 'v'))

    def test_rouine_init(self):
        """
        Test `Routine.init()` method.
        """

        self.assertTrue(self.ss.DCOPF.init(), "DCOPF initialization failed!")


@unittest.skipUnless(HAVE_IGRAPH, "igraph not available")
class TestRoutineGraph(unittest.TestCase):
    """
    Test routine graph.
    """

    def test_5bus_graph(self):
        """
        Test routine graph of PJM 5-bus system.
        """
        ss = ams.load(ams.get_case("5bus/pjm5bus_uced.xlsx"),
                      default_config=True,
                      no_output=True,
                      )
        _, g = ss.DCOPF.igraph()
        self.assertGreaterEqual(np.min(g.degree()), 1)

    def test_ieee14_graph(self):
        """
        Test routine graph of IEEE 14-bus system.
        """
        ss = ams.load(ams.get_case("ieee14/ieee14_uced.xlsx"),
                      default_config=True,
                      no_output=True,
                      )
        _, g = ss.DCOPF.igraph()
        self.assertGreaterEqual(np.min(g.degree()), 1)

    def test_ieee39_graph(self):
        """
        Test routine graph of IEEE 39-bus system.
        """
        ss = ams.load(ams.get_case("ieee39/ieee39_uced_esd1.xlsx"),
                      default_config=True,
                      no_output=True,
                      )
        _, g = ss.DCOPF.igraph()
        self.assertGreaterEqual(np.min(g.degree()), 1)

    def test_npcc_graph(self):
        """
        Test routine graph of NPCC 140-bus system.
        """
        ss = ams.load(ams.get_case("npcc/npcc_uced.xlsx"),
                      default_config=True,
                      no_output=True,
                      )
        _, g = ss.DCOPF.igraph()
        self.assertGreaterEqual(np.min(g.degree()), 1)

    def test_wecc_graph(self):
        """
        Test routine graph of WECC 179-bus system.
        """
        ss = ams.load(ams.get_case("wecc/wecc_uced.xlsx"),
                      default_config=True,
                      no_output=True,
                      )
        _, g = ss.DCOPF.igraph()
        self.assertGreaterEqual(np.min(g.degree()), 1)



"""
ACOPF routines.
"""
import logging
from collections import OrderedDict

from ams.pypower import runopf
from ams.pypower.core import ppoption

from ams.io.pypower import system2ppc
from ams.core.param import RParam

from ams.routines.dcpf import DCPFlowBase
from ams.opt.omodel import Var, Constraint, Objective

logger = logging.getLogger(__name__)


class ACOPFBase(DCPFlowBase):
    """
    Base class for ACOPF model.
    """

    def __init__(self, system, config):
        DCPFlowBase.__init__(self, system, config)
        self.map1 = OrderedDict()   # ACOPF does not receive
        self.map2 = OrderedDict([
            ('Bus', {
                'vBus': 'v0',
            }),
            ('StaticGen', {
                'pg': 'p0',
            }),
        ])

    def solve(self, method=None, **kwargs):
        """
        Solve ACOPF using PYPOWER with PIPS.
        """
        ppc = system2ppc(self.system)
        ppopt = ppoption()
        res, sstats = runopf(casedata=ppc, ppopt=ppopt, **kwargs)
        self.converged = res['success']
        return res, self.converged, sstats

    def unpack(self, res):
        """
        Unpack results from PYPOWER.
        """
        super().unpack(res)

        # --- Bus ---
        bus_idx = self.vBus.get_idx()
        self.vBus.optz.value = self.system.Bus.get(src='v', attr='v', idx=bus_idx)
        self.aBus.optz.value = self.system.Bus.get(src='a', attr='v', idx=bus_idx)

        # --- Gen ---
        gen_idx = self.pg.get_idx()
        self.pg.optz.value = self.system.StaticGen.get(src='p', attr='v', idx=gen_idx)
        self.qg.optz.value = self.system.StaticGen.get(src='q', attr='v', idx=gen_idx)

        # --- Objective ---
        self.obj.v = res['f']

        self.system.recent = self.system.routines[self.class_name]
        return True

    def run(self, force_init=False, no_code=True,
            method=None, **kwargs):
        """
        Run ACOPF using PYPOWER with PIPS.

        Examples
        --------
        >>> ss = ams.load(ams.get_case('matpower/case14.m'))
        >>> ss.ACOPF.run()

        Parameters
        ----------
        force_init : bool
            Force initialization.
        no_code : bool
            Disable showing code.
        method : str
            Placeholder for future use.

        Returns
        -------
        exit_code : int
            Exit code of the routine.
        """
        super().run(force_init=force_init,
                    no_code=no_code, method=method,
                    **kwargs, )


class ACOPF(ACOPFBase):
    """
    Standard AC optimal power flow.

    Notes
    -----
    1. ACOPF is solved with PYPOWER ``runopf`` function.
    2. ACOPF formulation in AMS style is NOT DONE YET,
       but this does not affect the results
       because the data are passed to PYPOWER for solving.
    """

    def __init__(self, system, config):
        ACOPFBase.__init__(self, system, config)
        self.info = 'AC Optimal Power Flow'
        self.type = 'ACED'

        # --- params ---
        self.c2 = RParam(info='Gen cost coefficient 2',
                         name='c2', tex_name=r'c_{2}',
                         unit=r'$/(p.u.^2)', model='GCost',
                         indexer='gen', imodel='StaticGen',
                         nonneg=True)
        self.c1 = RParam(info='Gen cost coefficient 1',
                         name='c1', tex_name=r'c_{1}',
                         unit=r'$/(p.u.)', model='GCost',
                         indexer='gen', imodel='StaticGen',)
        self.c0 = RParam(info='Gen cost coefficient 0',
                         name='c0', tex_name=r'c_{0}',
                         unit=r'$', model='GCost',
                         indexer='gen', imodel='StaticGen',
                         no_parse=True)
        self.qd = RParam(info='reactive demand',
                         name='qd', tex_name=r'q_{d}',
                         model='StaticLoad', src='q0',
                         unit='p.u.',)
        # --- bus ---
        self.aBus = Var(info='Bus voltage angle',
                        unit='rad',
                        name='aBus', tex_name=r'a_{Bus}',
                        model='Bus', src='a',)
        self.vBus = Var(info='Bus voltage magnitude',
                        unit='p.u.',
                        name='vBus', tex_name=r'v_{Bus}',
                        src='v', model='Bus',)
        # --- gen ---
        self.pg = Var(info='Gen active power',
                      unit='p.u.',
                      name='pg', tex_name=r'p_{g}',
                      model='StaticGen', src='p',)
        self.qg = Var(info='Gen reactive power',
                      unit='p.u.',
                      name='qg', tex_name=r'q_{g}',
                      model='StaticGen', src='q',)
        # --- constraints ---
        self.pb = Constraint(name='pb',
                             info='power balance',
                             e_str='sum(pd) - sum(pg)',
                             type='eq',)
        # TODO: ACOPF formulation
        # --- objective ---
        self.obj = Objective(name='obj',
                             info='total cost',
                             e_str='sum(c2 * pg**2 + c1 * pg + c0)',
                             sense='min',)



import unittest

import ams
import numpy as np

ams.config_logger(stream_level=40)


class TestAddressing(unittest.TestCase):
    """
    Tests for addressing.
    """

    def test_ieee14_address(self):
        """
        Test IEEE14 address.
        """

        # FIXME: why there will be case parsing information using ams.system.example()?
        ss = ams.system.example()

        # bus variable indices (internal)
        np.testing.assert_array_equal(ss.Bus.a.a,
                                      np.arange(0, ss.Bus.n, 1))
        np.testing.assert_array_equal(ss.Bus.v.a,
                                      np.arange(ss.Bus.n, 2*ss.Bus.n, 1))

        # external variable indices
        np.testing.assert_array_equal(ss.PV.ud.a,
                                      np.array([28, 29, 30, 31]))
        np.testing.assert_array_equal(ss.PV.p.a,
                                      np.array([32, 33, 34, 35]))
        np.testing.assert_array_equal(ss.PV.q.a,
                                      np.array([36, 37, 38, 39]))
        np.testing.assert_array_equal(ss.Slack.ud.a,
                                      np.array([40]))
        np.testing.assert_array_equal(ss.Slack.p.a,
                                      np.array([41]))
        np.testing.assert_array_equal(ss.Slack.q.a,
                                      np.array([42]))



from andes.models.line.line import LineData
from andes.core.param import NumParam
from andes.shared import deg2rad

from ams.core.model import Model


class Line(LineData, Model):
    """
    AC transmission line model.

    The model is also used for two-winding transformer. Transformers can set the
    tap ratio in ``tap`` and/or phase shift angle ``phi``.

    Notes
    -----
    There is a known issue that adding Algeb ``ud`` will cause Line.algebs run into
    AttributeError: 'NoneType' object has no attribute 'n'. Not figured out why yet.
    """

    def __init__(self, system=None, config=None) -> None:
        LineData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'ACLine'

        self.amin = NumParam(default=-360 * deg2rad,
                             info="minimum angle difference, from bus - to bus",
                             unit='rad',
                             tex_name=r'a_{min}',
                             )
        self.amax = NumParam(default=360 * deg2rad,
                             info="maximum angle difference, from bus - to bus",
                             unit='rad',
                             tex_name=r'a_{max}',
                             )
        self.rate_a.default = 999.0
        self.rate_b.default = 999.0
        self.rate_c.default = 999.0



"""
Shared constants and delayed imports.

This module is supplementary to the ``andes.shared`` module.
"""
import logging
from functools import wraps
from datetime import datetime

import cvxpy as cp

from andes.utils.lazyimport import LazyImport


logger = logging.getLogger(__name__)


igraph = LazyImport("import igraph")

# NOTE: copied from CVXPY documentation
MIP_SOLVERS = ['CBC', 'COPT', 'GLPK_MI', 'CPLEX', 'GUROBI',
               'MOSEK', 'SCIP', 'XPRESS', 'SCIPY']

INSTALLED_SOLVERS = cp.installed_solvers()

# NOTE: copyright
year_end = datetime.now().year
copyright_msg = f'Copyright (C) 2023-{year_end} Jinning Wang'


def require_MIP_solver(f):
    """
    Decorator for functions that require MIP solver.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if not any(s in MIP_SOLVERS for s in INSTALLED_SOLVERS):
            raise ModuleNotFoundError("No MIP solver is available.")
        return f(*args, **kwargs)

    return wrapper


def require_igraph(f):
    """
    Decorator for functions that require igraph.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            getattr(igraph, "__version__")
        except AttributeError:
            logger.error("Package `igraph` is not installed.")
        return f(*args, **kwargs)

    return wrapper



import unittest

import ams


class TestCodegen(unittest.TestCase):
    """
    Test code generation.
    """

    def test_1_docs(self) -> None:
        sp = ams.system.System()
        out = ''
        for tp in sp.types.values():
            out += tp.doc_all()



import logging
import unittest

import ams

logger = logging.getLogger(__name__)


class TestKnownResults(unittest.TestCase):
    # NOTE: DCOPF objective values are from MATPOWER v8.0b1
    sets = (
        (ams.get_case('matpower/case14.m'), 7642.59177699),
        (ams.get_case('matpower/case39.m'), 41263.94078588),
        (ams.get_case('matpower/case118.m'), 125947.8814179),
    )

    def test_known_results(self):
        for case, obj in self.sets:
            sp = ams.load(case, setup=True, no_output=True)
            sp.DCOPF.run(solver='ECOS')
            msg = f'Case: {case}; Result: {sp.DCOPF.obj.v}; Expected: {obj}'
            logger.info(msg)
            self.assertAlmostEqual(sp.DCOPF.obj.v, obj, places=2)



"""
Module for system matrix make.
"""

import logging
from typing import Optional

import numpy as np

from scipy.sparse import csr_matrix as c_sparse
from scipy.sparse import lil_matrix as l_sparse

from andes.utils.misc import elapsed

from ams.opt.omodel import Param

logger = logging.getLogger(__name__)


class MParam(Param):
    """
    Class for matrix parameters built from the system.

    MParam is designed to be a subclass of RParam for routine parameters
    management.

    Parameters
    ----------
    name : str, optional
        Name of this parameter. If not provided, `name` will be set
        to the attribute name.
    tex_name : str, optional
        LaTeX-formatted parameter name. If not provided, `tex_name`
        will be assigned the same as `name`.
    info : str, optional
        A description of this parameter
    unit : str, optional
        Unit of the parameter.
    v : np.ndarray, optional
        Matrix value of the parameter.
    owner : object, optional
        Owner of the MParam, usually the MatProcessor instance.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 v: Optional[np.ndarray] = None,
                 sparse: Optional[bool] = False,
                 ):
        self.name = name
        self.tex_name = tex_name if (tex_name is not None) else name
        self.info = info
        self.unit = unit
        self._v = v
        self.sparse = sparse
        self.owner = None

    @property
    def v(self):
        """
        Return the value of the parameter.
        """
        # NOTE: scipy.sparse matrix will return 2D array
        # so we squeeze it here if only one row
        if isinstance(self._v, (c_sparse, l_sparse)):
            out = self._v.toarray()
            if out.shape[0] == 1:
                return np.squeeze(out)
            else:
                return out
        return self._v

    @property
    def shape(self):
        """
        Return the shape of the parameter.
        """
        return self.v.shape

    @property
    def n(self):
        """
        Return the szie of the parameter.
        """
        return len(self.v)

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__


class MatProcessor:
    """
    Class for matrix processing in AMS system.
    """

    def __init__(self, system):
        self.system = system
        self.Bbus = MParam(name='Bbus', tex_name=r'B_{bus}',
                           info='Bus admittance matrix',
                           v=None, sparse=True)
        self.Bf = MParam(name='Bf', tex_name=r'B_{f}',
                         info='Bf matrix',
                         v=None, sparse=True)
        self.Pbusinj = MParam(name='Pbusinj', tex_name=r'P_{bus}^{inj}',
                              info='Bus power injection vector',
                              v=None,)
        self.Pfinj = MParam(name='Pfinj', tex_name=r'P_{f}^{inj}',
                            info='Line power injection vector',
                            v=None,)
        self.PTDF = MParam(name='PTDF', tex_name=r'P_{TDF}',
                           info='Power transfer distribution factor',
                           v=None)

        self.Cft = MParam(name='Cft', tex_name=r'C_{ft}',
                          info='Line connectivity matrix',
                          v=None, sparse=True)
        self.CftT = MParam(name='CftT', tex_name=r'C_{ft}^{T}',
                           info='Line connectivity matrix transpose',
                           v=None, sparse=True)
        self.Cg = MParam(name='Cg', tex_name=r'C_g',
                         info='Generator connectivity matrix',
                         v=None, sparse=True)
        self.Cs = MParam(name='Cs', tex_name=r'C_s',
                         info='Slack connectivity matrix',
                         v=None, sparse=True)
        self.Cl = MParam(name='Cl', tex_name=r'Cl',
                         info='Load connectivity matrix',
                         v=None, sparse=True)
        self.Csh = MParam(name='Csh', tex_name=r'C_{sh}',
                          info='Shunt connectivity matrix',
                          v=None, sparse=True)

    def make(self):
        """
        Make the system matrices.

        Note that this method will update all matrices in the class.
        """
        t_mat, _ = elapsed()
        self._makeC()
        self._makeBdc()
        _, s_mat = elapsed(t_mat)
        logger.debug(f"Built system matrices in {s_mat}.")

        return True

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    @property
    def n(self):
        """
        To fit the RParam style.
        """
        return 2

    def _makeBdc(self):
        """
        Make Bdc matrix.

        Call _makeC() before this method to ensure Cft is available.
        """
        system = self.system

        # common variables
        nb = system.Bus.n
        nl = system.Line.n

        # line parameters
        idx_line = system.Line.idx.v
        x = system.Line.get(src='x', attr='v', idx=idx_line)
        u_line = system.Line.get(src='u', attr='v', idx=idx_line)
        b = u_line / x  # series susceptance

        # in DC, tap is assumed to be 1
        tap0 = system.Line.get(src='tap', attr='v', idx=idx_line)
        tap = np.ones(nl)
        i = np.flatnonzero(tap0)
        tap[i] = tap0[i]  # assign non-zero tap ratios
        b = b / tap  # adjusted series susceptance

        # build Bf such that Bf * Va is the vector of real branch powers injected
        # at each branch's "from" bus
        f = system.Bus.idx2uid(system.Line.get(src='bus1', attr='v', idx=idx_line))
        t = system.Bus.idx2uid(system.Line.get(src='bus2', attr='v', idx=idx_line))
        ir = np.r_[range(nl), range(nl)]  # double set of row indices
        Bf = c_sparse((np.r_[b, -b], (ir, np.r_[f, t])), (nl, nb))

        # build Cft, note that this Cft is different from the one in _makeC()
        Cft = c_sparse((np.r_[np.ones(nl), -np.ones(nl)], (ir, np.r_[f, t])), (nl, nb))

        # build Bbus
        Bbus = Cft.T * Bf

        phi = system.Line.get(src='phi', attr='v', idx=idx_line)
        Pfinj = b * (-phi)
        Pbusinj = Cft.T * Pfinj

        self.Bbus._v, self.Bf._v, self.Pbusinj._v, self.Pfinj._v = Bbus, Bf, Pbusinj, Pfinj
        return True

    def _makeC(self):
        """
        Make connectivity matrix.
        """
        system = self.system

        # common variables
        nb = system.Bus.n
        ng = system.StaticGen.n
        nl = system.Line.n
        npq = system.PQ.n
        nsh = system.Shunt.n

        # --- Cg ---
        # bus indices: idx -> uid
        idx_gen = system.StaticGen.get_idx()
        u_gen = system.StaticGen.get(src='u', attr='v', idx=idx_gen)
        on_gen = np.flatnonzero(u_gen)  # uid of online generators
        on_gen_idx = [idx_gen[i] for i in on_gen]  # idx of online generators
        on_gen_bus_idx = system.StaticGen.get(src='bus', attr='v', idx=on_gen_idx)

        row = np.array([system.Bus.idx2uid(x) for x in on_gen_bus_idx])
        col = np.array([system.StaticGen.idx2uid(x) for x in on_gen_idx])
        Cg = c_sparse((np.ones(len(on_gen_idx)), (row, col)), (nb, ng))

        # --- Cl ---
        # load indices: idx -> uid
        idx_load = system.PQ.idx.v
        u_load = system.PQ.get(src='u', attr='v', idx=idx_load)
        on_load = np.flatnonzero(u_load)  # uid of online loads
        on_load_idx = [idx_load[i] for i in on_load]  # idx of online loads
        on_load_bus_idx = system.PQ.get(src='bus', attr='v', idx=on_load_idx)

        row = np.array([system.Bus.idx2uid(x) for x in on_load_bus_idx])
        col = np.array([system.PQ.idx2uid(x) for x in on_load_idx])
        Cl = c_sparse((np.ones(len(on_load_idx)), (row, col)), (nb, npq))

        # --- Cft ---
        # line indices: idx -> uid
        idx_line = system.Line.idx.v
        u_line = system.Line.get(src='u', attr='v', idx=idx_line)
        on_line = np.flatnonzero(u_line)  # uid of online lines
        on_line_idx = [idx_line[i] for i in on_line]  # idx of online lines
        on_line_bus1_idx = system.Line.get(src='bus1', attr='v', idx=on_line_idx)
        on_line_bus2_idx = system.Line.get(src='bus2', attr='v', idx=on_line_idx)

        data_line = np.ones(2*len(on_line_idx))
        data_line[len(on_line_idx):] = -1
        row_line = np.array([system.Bus.idx2uid(x) for x in on_line_bus1_idx + on_line_bus2_idx])
        col_line = np.array([system.Line.idx2uid(x) for x in on_line_idx + on_line_idx])
        Cft = c_sparse((data_line, (row_line, col_line)), (nb, nl))

        # --- Csh ---
        # shunt indices: idx -> uid
        idx_shunt = system.Shunt.idx.v
        u_shunt = system.Shunt.get(src='u', attr='v', idx=idx_shunt)
        on_shunt = np.flatnonzero(u_shunt)  # uid of online shunts
        on_shunt_idx = [idx_shunt[i] for i in on_shunt]  # idx of online shunts
        on_shunt_bus_idx = system.Shunt.get(src='bus', attr='v', idx=on_shunt_idx)

        row = np.array([system.Bus.idx2uid(x) for x in on_shunt_bus_idx])
        col = np.array([system.Shunt.idx2uid(x) for x in on_shunt_idx])
        Csh = c_sparse((np.ones(len(on_shunt_idx)), (row, col)), (nb, nsh))

        self.Cg._v, self.Cl._v, self.Cft._v, self.Csh._v = Cg, Cl, Cft, Csh
        self.CftT._v = Cft.T
        return True
